package lopsil;

import lopsil.device.*;
import javax.swing.JOptionPane;

/**
 * PolicyAssumptions sets up the devices and assumptions that are to be used for the policy.
 * It checks for at least one valid device so the application can continue.  It also starts the
 * FrequencyOfUpdatesAssumption thread if one is specified.
 * 
 * @author Billy Rickey
 * @author Jay Ligatti
 *
 */
public class PolicyAssumptions {

	/**
	 * Constructs a newly allocated PolicyAssumptions which sets up the policy.
	 * If the GPSGranularities do not comply, the Policy's <code>handleGranularityViolation()</code> will
	 * be called.
	 * 
	 * @param p the Policy to apply the assumption to
	 * @param LocationDevice An array of LocationDevices to be used in the Policy
	 * @param t The FrequencyOfUpdatesAssumption to be used in the Policy
	 * @param g The LocationGranularityAssumption to be used in the Policy
	 */
	public PolicyAssumptions(Policy p, LocationDevice[] locationDevice, LocationGranularityAssumption g, FrequencyOfUpdatesAssumption f)
	{
		Location.setPolicy(p);
		if(g.getValue(LocationGranularityAssumption.YARDS) != Units.INFINITY) {
			//If a LocationGranularityAssumption was specified, make sure a GPS LocationDevice is present
			try{
				checkGPS(locationDevice);
			} catch(Exception e) {
				try {
					e.printStackTrace();
					System.out.println("policyass");
					p.handleGranularityViolation();
				} catch (Exception geoEx) {
					if(JOptionPane.showConfirmDialog(null, geoEx.toString(), "Continue?", JOptionPane.YES_NO_OPTION) != 0)
						System.exit(1);
				}
			}
		}
		
		// Setup an array of Locations to be assigned to the LocationDevices
		Location[] l = new Location[locationDevice.length];

		for(int i = 0; i < locationDevice.length; i++) {
			l[i] = locationDevice[i].getLocation();
			//locationDevice[i].setPolicy(p);
			if(locationDevice[i].isGPS()) {
				// Check for GPSGranulairty compliance
				if(locationDevice[i].getGranularity().compliesWith(g)) {
					if(locationDevice[i].isActive()) 
						locationDevice[i].setValid();
				} else {
					try {
						System.out.println("policyass2");
						p.handleGranularityViolation();
					} catch (Exception geoEx) {
						if(JOptionPane.showConfirmDialog(null, geoEx.toString() + "\nProblem exists with Device " + i,
								"Set valid?", JOptionPane.YES_NO_OPTION) == 0)
							locationDevice[i].setValid();
					}
				}
			} else
				locationDevice[i].setValid();
		}
		try{
			checkValidity(locationDevice);			
			if(f != null) {
				Thread timeThread = new Thread(new FrequencyOfUpdatesAssumption(f, locationDevice, p));
				timeThread.setDaemon(true);
				timeThread.start();
			}
		} catch(Exception e) {
			if(JOptionPane.showConfirmDialog(null, e.toString(), "Continue?", JOptionPane.YES_NO_OPTION) != 0)
				System.exit(1);
		}
	}
	
	/**
	 * Checks to make sure that a GPS device was specified if a LocationGranularityAssumption was specified.
	 * 
	 * @param locationDevice The array of LocationDevices
	 * @throws LopsilException
	 */
	private void checkGPS(LocationDevice[] locationDevice) throws LopsilException {
		boolean exception = true;
		for(int i = 0; i < locationDevice.length; i++)
			if(locationDevice[i].isGPS())
				exception = false;
		if(exception)
			throw new LopsilException("You specified a LocationGranularityAssumption without specifying a GPS device!");
	}
	
	/**
	 * Checks to make sure that at least one valid device is available.
	 * 
	 * @param locationDevice The array of LocationDevices
	 * @throws LopsilException
	 */
	private void checkValidity(LocationDevice[] locationDevice) throws LopsilException {
		boolean exception = true;
		for(int i = 0; i < locationDevice.length; i++)
			if(locationDevice[i].isValid())
				exception = false;
		if(exception)
			throw new LopsilException("No valid LocationDevice detected in system!");
	}
	
} // end PolicyAssumptions
